from microdf.generic import MicroDataFrame
from openfisca_uk.data.datasets.was import WAS
from openfisca_uk.data.datasets.frs.frs import FRS
import pandas as pd
import microdf as mdf
import synthimpute as si

REGIONS = {
    1: "NORTH_EAST",
    2: "NORTH_WEST",
    3: "YORKSHIRE",
    4: "EAST_MIDLANDS",
    5: "WEST_MIDLANDS",
    6: "EAST_OF_ENGLAND",
    7: "LONDON",
    8: "SOUTH_EAST",
    9: "SOUTH_WEST",
    10: "WALES",
    11: "SCOTLAND",
    12: "NORTHERN_IRELAND",
}

PREDICTOR_VARIABLES = [
    "household_net_income",
    "num_adults",
    "num_children",
    "pension_income",
    "employment_income",
    "self_employment_income",
    "capital_income",
    "num_bedrooms",
    "council_tax",
    "is_renting",
    "region",
]

IMPUTE_VARIABLES = [
    "owned_land",
    "property_wealth",
    "corporate_wealth",
    "gross_financial_wealth",
    "net_financial_wealth",
    "main_residence_value",
    "other_residential_property_value",
    "non_residential_property_value",
]


def impute_wealth(year: int, dataset: type = FRS) -> pd.Series:
    """Impute wealth by fitting a random forest model.

    Args:
        year (int): The year of simulation.
        dataset (type): The dataset to use.

    Returns:
        pd.Series: The predicted wealth values.
    """

    from openfisca_uk import Microsimulation

    sender = Microsimulation(dataset=WAS, year=year,).df(
        PREDICTOR_VARIABLES + IMPUTE_VARIABLES,
        map_to="household",
    )

    receiver = Microsimulation(dataset=dataset, year=year,).df(
        PREDICTOR_VARIABLES,
        map_to="household",
    )

    sender.region = sender.region.map(
        {name: float(i) for i, name in REGIONS.items()}
    )
    receiver.region = receiver.region.map(
        {name: float(i) for i, name in REGIONS.items()}
    )

    return si.rf_impute(
        x_train=sender.drop(IMPUTE_VARIABLES, axis=1),
        y_train=sender[IMPUTE_VARIABLES],
        x_new=receiver,
        verbose=True,
    )



import shutil
from openfisca_tools.data import PrivateDataset
import pandas as pd
from pathlib import Path
from tqdm import tqdm
from openfisca_uk.data.storage import OPENFISCA_UK_MICRODATA_FOLDER


class RawSPI(PrivateDataset):
    name = "raw_spi"
    label = "Raw SPI"
    folder_path = OPENFISCA_UK_MICRODATA_FOLDER
    is_openfisca_compatible = False

    filename_by_year = {
        2018: "raw_spi_2018.h5",
    }

    def generate(self, year: int, ukds_tab_zipfile: str):
        """Generates the raw LCFS tabular dataset from the TAB zip archive
        downloadable from the UKDS.

        Args:
            year (int): The year of the FRS to generate.
            ukds_tab_zipfile (str): The path to the TAB zip archive.
        """

        folder = Path(ukds_tab_zipfile)
        year = str(year)
        if not folder.exists():
            raise FileNotFoundError("Invalid path supplied")
        new_folder = self.folder_path / "tmp"
        shutil.unpack_archive(folder, new_folder)
        folder = new_folder
        main_folder = next(folder.iterdir())
        with pd.HDFStore(RawSPI.file(year)) as file:
            if (main_folder / "tab").exists():
                data_folder = main_folder / "tab"
                data_files = list(data_folder.glob("*.tab"))
                task = tqdm(data_files, desc="Saving data tables")
                for filepath in task:
                    task.set_description(f"Saving {filepath.name}")
                    table_name = "main"
                    df = pd.read_csv(
                        filepath, delimiter="\t", low_memory=False
                    ).apply(pd.to_numeric, errors="coerce")
                    df.columns = df.columns.str.upper()
                    file[table_name] = df
            else:
                raise FileNotFoundError("Could not find any TAB files.")

        # Clean up tmp storage.

        tmp_folder = self.folder_path / "tmp"
        if tmp_folder.exists():
            shutil.rmtree(tmp_folder)


RawSPI = RawSPI()



from openfisca_uk.data.datasets.frs.enhanced.stages.calibration.losses.loss_category import (
    LossCategory,
)
import tensorflow as tf
from openfisca_uk.parameters import parameters


class FamiliesInTotal(LossCategory):
    label = "Families in total"
    parameter_folder = parameters.calibration.families.in_total
    weight = 1 / 20

    def get_loss_subcomponents(
        sim,
        household_weights,
        year,
    ):
        num_families = sim.calc("families", map_to="household").values
        model_population = tf.reduce_sum(household_weights * num_families)
        actual_population = FamiliesInTotal.parameter_folder(f"{year}-01-01")
        yield FamiliesInTotal.parameter_folder.name + "." + str(
            year
        ), model_population, actual_population

    def get_metrics():
        return [FamiliesInTotal.parameter_folder]

    def get_metric_names():
        return [
            FamiliesInTotal.parameter_folder.name + "." + str(year)
            for year in range(2022, 2027)
        ]



import logging
import numpy as np
from openfisca_tools.data import Dataset, PublicDataset
import pandas as pd
import h5py
from openfisca_uk.data.datasets.frs.frs import FRS
from openfisca_uk.data.storage import OPENFISCA_UK_MICRODATA_FOLDER


class SynthFRS(PublicDataset):
    name = "synth_frs"
    label = "Synthetic FRS"
    is_openfisca_compatible = True
    url_by_year = {
        2022: "https://github.com/PolicyEngine/openfisca-uk/releases/download/synth-frs-v1/synth_frs_2022.h5",
    }
    folder_path = OPENFISCA_UK_MICRODATA_FOLDER

    format = Dataset.ARRAYS

    def generate(self, year: int):
        if year not in FRS.years:
            logging.warn(f"FRS for {year} not found: generating.")
            FRS.generate(year)

        ID_COLS = (
            "person_id",
            "person_benunit_id",
            "person_benunit_role",
            "person_household_id",
            "person_household_role",
            "person_state_id",
            "person_state_role",
            "benunit_id",
            "household_id",
            "state_id",
        )

        def anonymise(arr: np.array, name: str) -> pd.DataFrame:
            result = pd.Series(arr)
            if name not in ID_COLS:
                # don't change identity columns, this breaks structures
                if len(result.unique()) < 16:
                    # shuffle categorical columns
                    result = result.sample(frac=1).values
                else:
                    # shuffle + add noise to numeric columns
                    # noise = between -3% and +3% added to each row
                    noise = np.random.rand() * 3e-2 + 1.0
                    result = result.sample(frac=1).values * noise
            return result

        data = FRS.load(year)

        with h5py.File(SynthFRS.file(year), mode="w") as f:
            for variable in data.keys():
                try:
                    f[variable] = anonymise(data[variable], variable)
                except:
                    f[variable] = anonymise(data[variable], variable).astype(
                        "S"
                    )


SynthFRS = SynthFRS()



from .by_program_participation import (
    FamiliesByProgramParticipation,
)
from .in_total import (
    FamiliesInTotal,
)
from openfisca_uk.data.datasets.frs.enhanced.stages.calibration.losses.loss_category import (
    combine_loss_categories,
)


Families = combine_loss_categories(
    FamiliesByProgramParticipation, FamiliesInTotal, label="Families"
)



from openfisca_uk.data.datasets.frs.enhanced.stages.calibration.losses.loss_category import (
    LossCategory,
)
import tensorflow as tf
from openfisca_uk.parameters import parameters


class HouseholdsByRegionByTenureType(LossCategory):
    label = "Households by region and tenure type"
    parameter_folder = (
        parameters.calibration.households.by_region_by_tenure_type
    )
    weight = 1 / 12

    def get_loss_subcomponents(
        sim,
        household_weights,
        year,
    ):
        total_households = parameters.calibration.households.in_total(
            f"{year}-01-01"
        )
        total_tenure_metrics = sum(
            [
                parameter(f"{year}-01-01")
                for parameter in HouseholdsByRegionByTenureType.get_metrics()
            ]
        )
        adjustment = total_households / total_tenure_metrics
        tenure_type = sim.calc("ons_tenure_type").values
        hh_country = sim.calc("country").values
        tenure_metrics = HouseholdsByRegionByTenureType.parameter_folder
        for country in tenure_metrics.children:
            for tenure in tenure_metrics.children[country].children:
                parameter = tenure_metrics.children[country].children[tenure]
                parameter_name = parameter.name + "." + str(year)
                model_population = tf.reduce_sum(
                    (hh_country == country)
                    * (tenure_type == tenure)
                    * household_weights
                )
                actual_population = parameter(f"{year}-01-01") * adjustment
                yield parameter_name, model_population, actual_population



from openfisca_core.model_api import *
from openfisca_uk.entities import *
from openfisca_uk.tools.general import *
from openfisca_uk.tools.baseline_variables import (
    change_over_baseline,
    baseline_is_nonzero,
)



from openfisca_uk.data.datasets.frs.enhanced.stages.calibration.losses.loss_category import (
    LossCategory,
)
import tensorflow as tf
from openfisca_uk import Microsimulation
from typing import Iterable, Tuple
from openfisca_uk.parameters import parameters


class BudgetaryImpactByIncomeSource(LossCategory):
    label = "Budgetary impact by income source"
    parameter_folder = parameters.calibration.budgetary_impact.by_income_source
    weight = 1 / 10

    def get_loss_subcomponents(
        sim: Microsimulation, household_weights: tf.Tensor, year: int
    ) -> Iterable[Tuple]:
        aggregates = BudgetaryImpactByIncomeSource.parameter_folder
        variables = aggregates.children
        for variable in variables:
            values = sim.calc(variable, period=year).values
            entity = sim.simulation.tax_benefit_system.variables[
                variable
            ].entity.key
            household_totals = sim.map_to(values, entity, "household")
            # Calculate aggregate error
            agg = tf.reduce_sum(household_weights * household_totals)
            parameter = aggregates.children[variable]
            actual = parameter(f"{year}-01-01")
            yield parameter.name + "." + str(year), agg, actual



import numpy as np
import tensorflow as tf
from openfisca_uk.parameters import parameters
from openfisca_uk.data.datasets.frs.enhanced.stages.calibration.losses.loss_category import (
    LossCategory,
)


class PopulationsByRegion(LossCategory):
    label = "Population by region"
    parameter_folder = parameters.calibration.populations.by_region
    weight = 1 / 20

    def get_loss_subcomponents(
        sim,
        household_weights,
        year,
    ):
        household_region = sim.calc("region").values
        household_population = sim.calc("people", map_to="household").values
        regions = PopulationsByRegion.parameter_folder.children
        for region in regions:
            in_region = household_region == region
            people_in_region = household_population * in_region
            population = tf.reduce_sum(people_in_region * household_weights)
            target = regions[region](f"{year}-01-01")
            yield regions[region].name + "." + str(year), population, target



from typing import Callable, List
from openfisca_uk.api import *
from openfisca_uk.situation_examples.examples import single_person_UC
import pandas as pd


def get_wide_reform_population_data(
    reforms: List[Reform],
    names: List[str],
    variables: List[str] = [],
    mapping: str = "person",
    function: Callable = None,
    **kwargs,
) -> pd.DataFrame:
    """Generates wide-form data from given reforms and variable names
    (e.g. for creating animation frames for Plotly graphs).

    Args:
        reforms (List[Reform]): A list of Reform classes.
        names (List[str]): A list of variable names to calculate for each reform
        variables (List[str], optional): A list of variable names to calculate of each reform.
        mapping (person): The mapping to apply to each calculation (person, benunit, household).
        function (Callable[(baseline, reform_sim) -> pd.DataFrame]): A function to apply instead of calculating variables.
        kwargs: Keywork arguments passed to Microsimulations.
    Returns:
        pd.DataFrame: A DataFrame with each row containing:
        {reform: name, baseline_first_variable_name: [...],
        reform_first_variable_name: [...], ...}
    """
    dfs = []
    baseline = Microsimulation(**kwargs)
    for reform, name in zip(reforms, names):
        reform_sim = Microsimulation(reform, **kwargs)
        df = pd.DataFrame()
        if function is None:
            for variable in variables:
                df[f"reform_{variable}"] = reform_sim.calc(
                    variable, map_to=mapping
                )
                df[f"baseline_{variable}"] = baseline.calc(
                    variable, map_to=mapping
                )
                df["Reform"] = name
                df["weight"] = baseline.calc(f"{mapping}_weight")
        else:
            df = function(baseline, reform_sim)
            df["Reform"] = name
        dfs += [df]
    df = pd.concat(dfs)
    return df


def get_wide_reform_individual_data(
    reforms: List[Reform],
    names: List[str],
    variables: List[str] = ["net_income"],
    situation_function: Callable = single_person_UC,
    varying: str = "employment_income",
    vary_min: float = 0,
    vary_max: float = 200000,
    vary_step: float = 100,
    include_derivatives: bool = True,
    primary_adult_name: str = "adult",
    **kwargs,
):
    """Generates wide-format individual data from a set of reforms, varying a
    variable (a row for each reform-input value-other values set).

    Args:
        reforms (List[Reform]): A list of reforms.
        names (List[str]): The corresponding reform names.
        variables (List[str], optional): The variables to include. Defaults to ["net_income"].
        situation_function (Callable, optional): A function adding situational data to the simulation.
            Defaults to single_person_UC.
        varying (str, optional): The variable to vary (e.g. employment income). Defaults to
            "employment_income".
        vary_min (float, optional): The starting value of the varying variable. Defaults to 0.
        vary_max (float, optional): The ending value of the varying variable. Defaults to 200,000.
        vary_step (float, optional): The interval of the varying variable. Defaults to 100.
        include_derivatives (bool, optional): Whether to include derivatives for each
            calculation. Defaults to True.
        primary_adult_name (str, optional): The name of the adult to calculate derivatives
            for. Defaults to "adult".
        kwargs: Keyword arguments passed to Microsimulations.

    Returns:
        pd.DataFrame: The resulting dataframe.
    """
    baseline = IndividualSim(**kwargs)
    baseline = situation_function(baseline)
    baseline.vary(varying)
    dfs = []
    for reform, name in zip(reforms, names):
        df = pd.DataFrame()
        reform_sim = IndividualSim(reform, **kwargs)
        reform_sim = situation_function(reform_sim)
        reform_sim.vary(varying, min=vary_min, max=vary_max, step=vary_step)
        df[varying] = baseline.calc(varying)[0]
        for variable in variables:
            try:
                df[f"baseline_{variable}"] = baseline.calc(variable).sum(
                    axis=0
                )
                if include_derivatives:
                    df[f"baseline_{variable}_deriv"] = baseline.deriv(
                        variable,
                        wrt=varying,
                        var_target=primary_adult_name,
                        wrt_target=primary_adult_name,
                    )
            except:
                df[f"baseline_{variable}"] = np.zeros_like(df[varying])
                if include_derivatives:
                    df[f"baseline_{variable}_deriv"] = np.zeros_like(
                        df[varying]
                    )
            df[f"reform_{variable}"] = reform_sim.calc(variable).sum(axis=0)
            if include_derivatives:
                df[f"reform_{variable}_deriv"] = reform_sim.deriv(
                    variable,
                    wrt=varying,
                    var_target=primary_adult_name,
                    wrt_target=primary_adult_name,
                )
            df[f"reform"] = name
        dfs += [df]
    df = pd.concat(dfs)
    return df



import numpy as np

from openfisca_uk.data.datasets.frs.enhanced.stages.calibration.losses.loss_category import (
    LossCategory,
)
import tensorflow as tf
from openfisca_uk.parameters import parameters
from openfisca_core.parameters import Parameter


class PopulationsByAgeBand(LossCategory):
    label = "Populations by age band"
    parameter_folder = parameters.calibration.populations.by_age_band
    weight = 1 / 20

    def get_loss_subcomponents(
        sim,
        household_weights,
        year,
    ):
        total_population = parameters.calibration.populations.in_total(
            f"{year}-01-01"
        )
        total_age_bands = sum(
            [
                parameter(f"{year}-01-01")
                for parameter in PopulationsByAgeBand.parameter_folder.get_descendants()
                if isinstance(parameter, Parameter)
            ]
        )
        adjustment = total_population / total_age_bands
        population = PopulationsByAgeBand.parameter_folder
        person_age = sim.calc("age", year)
        for age_group in population.children:
            if "BETWEEN" in age_group:
                _, lower, upper = age_group.split("_")
                lower, upper = int(lower), int(upper)
            elif "OVER" in age_group:
                lower, upper = int(age_group.split("_")[1]) - 1, np.inf
            else:
                raise ValueError(f"Unexpected test group: {age_group}")
            people_in_household = sim.map_to(
                (person_age >= lower) * (person_age < upper),
                "person",
                "household",
            )
            model_population = tf.reduce_sum(
                people_in_household * household_weights
            )
            yield getattr(
                PopulationsByAgeBand.parameter_folder,
                age_group,
            ).name + "." + str(year), model_population, population.children[
                age_group
            ](
                f"{year}-01-01"
            ) * adjustment

    def get_metrics():
        return PopulationsByAgeBand.parameter_folder.get_descendants()

    def get_metric_names():
        return [
            parameter.name + "." + str(year)
            for year in range(2022, 2027)
            for parameter in PopulationsByAgeBand.get_metrics()
        ]



import numpy as np
from openfisca_core.parameters import ParameterNode, Parameter
from numpy.typing import ArrayLike
from itertools import chain
from functools import reduce
from typing import Callable, Iterable, List, Tuple, Type
from openfisca_uk import Microsimulation
import tensorflow as tf


def weighted_squared_relative_deviation(
    pred: tf.Tensor, actual: ArrayLike
) -> tf.Tensor:
    """Computes the squared relative deviation between two tensors, weighted by the actual value.

    Args:
        pred (tf.Tensor): Predicted values.
        actual (ArrayLike): Actual values.

    Returns:
        tf.Tensor: The weighted squared relative deviation.
    """
    return (pred - actual) ** 2


def weighted_squared_log_relative_deviation(
    pred: tf.Tensor, actual: ArrayLike
) -> tf.Tensor:
    """Computes the squared log relative deviation between two tensors, weighted by the actual value.

    Args:
        pred (tf.Tensor): Predicted values.
        actual (ArrayLike): Actual values.

    Returns:
        tf.Tensor: The weighted squared log relative deviation.
    """
    if actual == 0:
        return tf.constant(0, dtype=tf.float32)
    return (tf.math.log(pred / actual) - 1) ** 2 * actual


class LossCategory:
    weight: float = 1
    label: str
    category: str
    parameter_folder: ParameterNode
    years: List[int] = list(range(2022, 2027))
    comparison_loss_function: Callable = weighted_squared_relative_deviation
    initial_train_loss: float = None
    initial_val_loss: float = None
    cache = {}
    use_cache = False

    def get_loss_subcomponents(
        sim: Microsimulation, household_weights: tf.Tensor, year: int
    ) -> Iterable[Tuple]:
        """Gathers predictions used to measure weight performance.

        Args:
            sim (Microsimulation): A microsimulation from which to draw demographic data.
            household_weights (tf.Tensor): Household weights for each year.
            year (int): The year to test on.

        Returns:
            Iterable[Tuple]: A list of (name, pred, actual) triplets.
        """
        raise NotImplementedError()

    @classmethod
    def compute(
        cls,
        sim: Microsimulation,
        household_weights: tf.Tensor,
        excluded_metrics: List[str],
        validation: bool = False,
    ) -> Tuple[tf.Tensor, dict]:
        loss = tf.constant(0, dtype=tf.float32)
        log = []
        for year in cls.years:
            for name, pred, actual in cls.get_loss_subcomponents(
                sim, household_weights[year - cls.years[0]], year
            ):

                if name not in excluded_metrics:
                    l = cls.comparison_loss_function(pred, actual)
                    log += [
                        dict(
                            name=name,
                            pred=float(pred.numpy()),
                            actual=float(actual),
                            loss=float(l.numpy()),
                            category=cls.category,
                        )
                    ]
                    loss += l
        if not validation and cls.initial_train_loss is None:
            cls.initial_train_loss = loss.numpy()
        elif validation and cls.initial_val_loss is None:
            cls.initial_val_loss = loss.numpy()
        initial_value = (
            cls.initial_val_loss if validation else cls.initial_train_loss
        ) / cls.weight
        if initial_value == 0:
            initial_value += 1
        for entry in log:
            entry["loss"] /= initial_value
        if len(log) == 0:
            return loss, log
        return loss / initial_value, log

    @classmethod
    def get_metrics(cls) -> List[Parameter]:
        return list(
            filter(
                lambda p: isinstance(p, Parameter),
                cls.parameter_folder.get_descendants(),
            )
        )

    @classmethod
    def get_metric_names(cls) -> List[str]:
        for parameter in cls.parameter_folder.get_descendants():
            if isinstance(parameter, Parameter):
                yield from [f"{parameter.name}.{year}" for year in cls.years]


def combine_two_loss_categories(
    cls: Type[LossCategory],
    other_cls: Type[LossCategory],
    group_name: str = None,
) -> Type[LossCategory]:
    combined = type(
        f"{cls.__name__}+{other_cls.__name__}",
        (LossCategory,),
        {
            "weight": cls.weight + other_cls.weight,
            "label": f"{cls.label}+{other_cls.label}",
            "category": group_name or f"{cls.label}+{other_cls.label}",
            "get_loss_subcomponents": lambda sim, household_weights, year: chain(
                cls.get_loss_subcomponents(sim, household_weights, year),
                other_cls.get_loss_subcomponents(sim, household_weights, year),
            ),
            "get_metrics": lambda: chain(
                cls.get_metrics(), other_cls.get_metrics()
            ),
            "get_metric_names": lambda: chain(
                cls.get_metric_names(), other_cls.get_metric_names()
            ),
        },
    )

    @classmethod
    def combined_compute(new_class, *args, **kwargs):
        l_1, d_1 = cls.compute(*args, **kwargs)
        l_2, d_2 = other_cls.compute(*args, **kwargs)
        return l_1 + l_2, d_1 + d_2

    combined.compute = combined_compute
    return combined


def combine_loss_categories(
    *loss_categories: Type[LossCategory],
    weight: float = None,
    label: str = None,
) -> Type[LossCategory]:
    total_subcategory_weight = sum(
        [category.weight for category in loss_categories]
    )
    weight = weight or total_subcategory_weight
    for category in loss_categories:
        if not hasattr(category, "category"):
            category.category = label
        category.weight *= weight / total_subcategory_weight
    category = reduce(
        lambda cat_1, cat_2: combine_two_loss_categories(
            cat_1, cat_2, group_name=label
        ),
        loss_categories,
    )
    category.weight = weight or category.weight
    return category



import numpy as np
from openfisca_uk.data.datasets.frs.enhanced.stages.calibration.losses.loss_category import (
    LossCategory,
)
import tensorflow as tf
from openfisca_uk.parameters import parameters


class PopulationsByIncomeTaxBand(LossCategory):
    label = "Income tax payers by income"
    parameter_folder = parameters.calibration.populations.by_income_tax_band
    weight = 1 / 20

    def get_loss_subcomponents(
        sim,
        household_weights,
        year,
    ):
        tax_band = sim.calc("tax_band", period=year).values
        pays_basic_rate = (
            (tax_band == "BASIC")
            | (tax_band == "INTERMEDIATE")
            | (tax_band == "STARTER")
        )
        pays_higher_rate = tax_band == "HIGHER"
        pays_add_rate = tax_band == "ADDITIONAL"

        basic_rate_payers = sim.map_to(pays_basic_rate, "person", "household")
        higher_rate_payers = sim.map_to(
            pays_higher_rate, "person", "household"
        )
        add_rate_payers = sim.map_to(pays_add_rate, "person", "household")
        taxpayers = PopulationsByIncomeTaxBand.parameter_folder

        for value, target in zip(
            [
                basic_rate_payers,
                higher_rate_payers,
                add_rate_payers,
            ],
            [
                taxpayers.BASIC,
                taxpayers.HIGHER,
                taxpayers.ADDITIONAL,
            ],
        ):
            population = tf.reduce_sum(household_weights * value)
            yield target.name + "." + str(year), population, target(
                f"{year}-01-01"
            )



import logging
from openfisca_tools.data import Dataset, PrivateDataset
import h5py
from openfisca_uk.data.datasets.was.raw_was import RawWAS
from openfisca_uk.data.storage import OPENFISCA_UK_MICRODATA_FOLDER

RENAMES = {
    "R7xshhwgt": "household_weight",
    # Components for estimating land holdings.
    "DVLUKValR7_sum": "owned_land",  # In the UK.
    "DVPropertyR7": "property_wealth",
    "DVFESHARESR7_aggr": "emp_shares_options",
    "DVFShUKVR7_aggr": "uk_shares",
    "DVIISAVR7_aggr": "investment_isas",
    "DVFCollVR7_aggr": "unit_investment_trusts",
    "TotpenR7_aggr": "pensions",
    "DvvalDBTR7_aggr": "db_pensions",
    # Predictors for fusing to FRS.
    "dvtotgirR7": "gross_income",
    "NumAdultW7": "num_adults",
    "NumCh18W7": "num_children",
    # Household Gross Annual income from occupational or private pensions
    "DVGIPPENR7_AGGR": "pension_income",
    "DVGISER7_AGGR": "self_employment_income",
    # Household Gross annual income from investments
    "DVGIINVR7_aggr": "capital_income",
    # Household Total Annual Gross employee income
    "DVGIEMPR7_AGGR": "employment_income",
    "HBedrmW7": "num_bedrooms",
    "GORR7": "region",
    "DVPriRntW7": "is_renter",  # {1, 2} TODO: Get codebook values.
    "CTAmtW7": "council_tax",
    # Other columns for reference.
    "DVLOSValR7_sum": "non_uk_land",
    "HFINWNTR7_Sum": "net_financial_wealth",
    "DVLUKDebtR7_sum": "uk_land_debt",
    "HFINWR7_Sum": "gross_financial_wealth",
    "TotWlthR7": "wealth",
    "DVhvalueR7": "main_residence_value",
    "DVHseValR7_sum": "other_residential_property_value",
    "DVBlDValR7_sum": "non_residential_property_value",
    "DVTotinc_bhcR7": "household_net_income",
}

VARIABLES = [
    "household_net_income",
    "num_adults",
    "num_children",
    "pension_income",
    "employment_income",
    "self_employment_income",
    "capital_income",
    "num_bedrooms",
    "council_tax",
    "is_renting",
    "owned_land",
    "property_wealth",
    "corporate_wealth",
    "gross_financial_wealth",
    "net_financial_wealth",
    "main_residence_value",
    "other_residential_property_value",
    "non_residential_property_value",
]

RENAMES = {key.lower(): value for key, value in RENAMES.items()}


class WAS(PrivateDataset):
    name = "was"
    label = "WAS"
    data_format = Dataset.ARRAYS
    folder_path = OPENFISCA_UK_MICRODATA_FOLDER

    is_openfisca_compatible = True

    def generate(self, year: int):
        if year in self.years:
            self.remove(year)
        # Load raw FRS tables
        year = int(year)

        if len(RawWAS.years) == 0:
            raise FileNotFoundError(
                "Raw WAS not found. Please run `openfisca-uk data was generate [year]` first."
            )

        if year > max(RawWAS.years):
            logging.warn("Uprating a previous version of the WAS.")
            if len(self.years) == 0:
                self.generate(max(RawWAS.years))
            if len(self.years) > 0:
                lcfs_year = max(self.years)
                from openfisca_uk import Microsimulation

                sim = Microsimulation(dataset=self, year=max(self.years))
                lcfs = h5py.File(self.file(year), mode="w")
                for variable in self.keys(lcfs_year):
                    lcfs[variable] = sim.calc(variable).values
                lcfs.close()
                return

        assert (
            year == 2019
        ), "Currently, only the 2019 WAS is supported (table name changes not yet implemented)"
        # The 2019 year is actually over the two-year period from 2018 to 2020, but is usually referred to as 2019.

        # TODO: Handle different WAS releases

        was = RawWAS.load(year, "was_round_7_hhold_eul_jan_2022")

        was = was.rename(columns={col: col.lower() for col in was.columns})

        to_remove = []
        to_add = {}

        for key in RENAMES:
            old_key = str(key)
            if key not in was.columns:
                key = key.replace("r", "w")
            if key not in was.columns:
                key = key.replace("w", "r")
            if key not in was.columns:
                raise ValueError(f"Could not find column {key}")
            else:
                to_add[key] = RENAMES[old_key]
                to_remove.append(old_key)

        for key in to_remove:
            del RENAMES[key]

        for key in to_add:
            RENAMES[key] = to_add[key]

        was = was.rename(columns=RENAMES).fillna(0)[list(RENAMES.values())]

        was["is_renting"] = was["is_renter"] == 1

        was.household_net_income *= 52  # WAS uses monthly income

        was["non_db_pensions"] = was.pensions - was.db_pensions
        was["corporate_wealth"] = was[
            [
                "non_db_pensions",
                "emp_shares_options",
                "uk_shares",
                "investment_isas",
                "unit_investment_trusts",
            ]
        ].sum(axis=1)

        entity_index = (
            was.index.values
        )  # One-person households for simplicity for now

        with h5py.File(self.file(year), mode="w") as f:
            for entity_id_var in [
                "person_id",
                "benunit_id",
                "household_id",
                "person_benunit_id",
                "person_household_id",
            ]:
                f[entity_id_var] = entity_index

            f["person_benunit_role"] = ["adult"] * len(entity_index)
            f["person_household_role"] = ["adult"] * len(entity_index)
            f["person_state_id"] = [1] * len(entity_index)
            f["state_id"] = [1]

            for variable in VARIABLES + ["household_weight"]:
                f[variable] = was[variable].values


WAS = WAS()



from typing import Dict
import h5py
from numpy.typing import ArrayLike
import numpy as np
from openfisca_tools.data.dataset import Dataset


def clone_and_replace_half(
    dataset: Dataset,
    year: int,
    mapping: Dict[str, ArrayLike],
    weighting: float = 0.5,
    target_dataset: Dataset = None,
):
    """Clones a dataset and replaces half of the values with the values in the mapping.

    Args:
        dataset (Dataset): The dataset to clone.
        year (int): The year to clone.
        mapping (Dict[str, ArrayLike]): The mapping from the original dataset to the cloned dataset.
        weighting (float): The weighting to apply to the cloned dataset. The original dataset has
                            (1 - weighting) weight.
        target_dataset (Dataset, optional): The dataset to write to. Defaults to None.
    """
    if target_dataset is None:
        target_dataset = dataset
    data = dataset.load(year)
    previous_data = {}
    for variable in data.keys():
        for period in data[variable].keys():
            previous_data[f"{variable}/{period}"] = data[variable][period][...]
    data.close()
    file = h5py.File(target_dataset.file(year), "w")
    for field in previous_data:
        if "_id" in field and not "state" in field:
            values = np.array(
                list(previous_data[field] * 10)
                + [x * 10 + 1 for x in previous_data[field]]
            )
        elif "_weight" in field and not "state" in field:
            values = np.concatenate(
                [
                    previous_data[field] * (1 - weighting),
                    previous_data[field] * weighting,
                ]
            )
        elif field in ("state_id", "state_weight"):
            values = previous_data[field]
        elif field in mapping:
            assert len(previous_data[field]) == len(
                mapping[field]
            ), f"Lengths don't match for {field}: {len(previous_data[field])} != {len(mapping[field])}"
            values = np.concatenate([previous_data[field], mapping[field]])
        else:
            values = np.concatenate(
                [previous_data[field], previous_data[field]]
            )
        try:
            file[field] = values
        except TypeError:
            file[field] = values.astype("S")
    file.close()


def add_variables(
    dataset: Dataset, year: int, variables: Dict[str, ArrayLike]
):
    data = dataset.load(year)
    previous_data = {}
    for variable in data.keys():
        for period in data[variable].keys():
            previous_data[f"{variable}/{period}"] = data[variable][period][...]
    data.close()
    f = h5py.File(dataset.file(year), "w")
    for field in previous_data:
        try:
            f[field] = previous_data[field]
        except TypeError:
            f[field] = previous_data[field].astype("S")
    for field in variables:
        try:
            f[field] = variables[field]
        except TypeError:
            f[field] = variables[field].astype("S")
    f.close()



from .in_total import (
    HouseholdsInTotal,
)
from .by_region_by_council_tax_band import (
    HouseholdsByRegionByCouncilTaxBand,
)
from .by_region_by_tenure_type import (
    HouseholdsByRegionByTenureType,
)
from openfisca_uk.data.datasets.frs.enhanced.stages.calibration.losses.loss_category import (
    combine_loss_categories,
)


Households = combine_loss_categories(
    HouseholdsInTotal,
    HouseholdsByRegionByCouncilTaxBand,
    HouseholdsByRegionByTenureType,
    label="Households",
)



from openfisca_uk.data.datasets.frs.enhanced.stages.calibration.losses.categories.populations.by_age_band import (
    PopulationsByAgeBand,
)
from .by_age_sex_region import PopulationsByAgeSexRegion
from .by_income_tax_band import (
    PopulationsByIncomeTaxBand,
)
from .by_region import (
    PopulationsByRegion,
)
from .in_total import (
    PopulationsInTotal,
)
from openfisca_uk.data.datasets.frs.enhanced.stages.calibration.losses.loss_category import (
    combine_loss_categories,
)


Populations = combine_loss_categories(
    PopulationsByAgeBand,
    PopulationsByIncomeTaxBand,
    PopulationsInTotal,
    label="Populations",
)



import logging
from openfisca_uk.data.datasets.spi.raw_spi import RawSPI
import pandas as pd
from pandas import DataFrame
import h5py
import numpy as np
from openfisca_tools.data import PrivateDataset

from openfisca_uk.data.storage import OPENFISCA_UK_MICRODATA_FOLDER

max_ = np.maximum
where = np.where


class SPI(PrivateDataset):
    name = "spi"
    label = "SPI"
    is_openfisca_compatible = True
    folder_path = OPENFISCA_UK_MICRODATA_FOLDER

    def generate(self, year: int) -> None:
        """Generates the SPI-based input dataset for OpenFisca-UK.

        Args:
            year (int): The year to generate for (uses the raw SPI from this year).
        """

        if year in self.years:
            self.remove(year)
        # Load raw FRS tables
        year = int(year)

        if len(RawSPI.years) == 0:
            raise FileNotFoundError(
                "Raw SPI not found. Please run `openfisca-uk data raw_spi generate [year]` first."
            )

        if year > max(RawSPI.years):
            logging.warn("Uprating a previous version of the SPI.")
            if len(self.years) == 0:
                self.generate(max(RawSPI.years))
            if len(self.years) > 0:
                frs_year = max(self.years)
                from openfisca_uk import Microsimulation

                sim = Microsimulation(dataset=self, year=max(self.years))
                frs = h5py.File(self.file(year), mode="w")
                for variable in self.keys(frs_year):
                    frs[variable] = sim.calc(variable).values
                frs.close()
                return

        main = RawSPI.load(year, "main").fillna(0)
        spi = h5py.File(self.file(year), mode="w")

        main = extend_spi_main_table(main)

        add_id_variables(spi, main)
        add_demographics(spi, main)
        add_incomes(spi, main)

        # Generate OpenFisca-UK variables and save
        spi.close()


SPI = SPI()


def extend_spi_main_table(main: DataFrame) -> DataFrame:
    """Extends the main SPI table to include adults and children with
    zero income, so that the total number of people is the UK population.

    Args:
        main (DataFrame): The main SPI table.

    Returns:
        DataFrame: The modified table.
    """

    from openfisca_uk import Microsimulation
    from openfisca_uk.data import FRS

    sim = Microsimulation(dataset=FRS)

    population_in_spi_percentage = main.FACT.sum() / sim.calc("people").sum()

    RENAMES = dict(
        pension_income="PENSION",
        self_employment_income="PROFITS",
        property_income="INCPROP",
        savings_interest_income="INCBBS",
        dividend_income="DIVIDENDS",
        blind_persons_allowance="BPADUE",
        married_couples_allowance="MCAS",
        gift_aid="GIFTAID",
        capital_allowances="CAPALL",
        deficiency_relief="DEFICIEN",
        covenanted_payments="COVNTS",
        charitable_investment_gifts="GIFTINV",
        employment_expenses="EPB",
        other_deductions="MOTHDED",
        pension_contributions="PENSRLF",
        employment_income="PAY",
        state_pension="SRP",
        miscellaneous_income="OTHERINC",
        pays_scottish_income_tax="SCOT_TXP",
        person_weight="FACT",
    )

    in_frs_and_not_spi = (
        sim.calc("total_income").rank(pct=True)
        < 1 - population_in_spi_percentage
    )

    missing_spi = sim.df(list(RENAMES))[in_frs_and_not_spi].rename(
        columns=RENAMES
    )

    LOWER = np.array([0, 16, 25, 35, 45, 55, 65, 75])
    UPPER = np.array([16, 25, 35, 45, 55, 65, 75, 80])
    CODE = np.array([1, 2, 3, 4, 5, 6, 7])
    REGIONS = {
        1: "NORTH_EAST",
        2: "NORTH_WEST",
        3: "YORKSHIRE",
        4: "EAST_MIDLANDS",
        5: "WEST_MIDLANDS",
        6: "EAST_OF_ENGLAND",
        7: "LONDON",
        8: "SOUTH_EAST",
        9: "SOUTH_WEST",
        10: "WALES",
        11: "SCOTLAND",
        12: "NORTHERN_IRELAND",
    }
    age = sim.calc("age")[in_frs_and_not_spi]
    missing_spi["AGERANGE"] = 0
    for lower, upper, code in zip(LOWER, UPPER, CODE):
        missing_spi["AGERANGE"] += np.where(
            (age < upper) & (age >= lower), code, 0
        )

    missing_spi["GORCODE"] = sim.calc("region", map_to="person")[
        in_frs_and_not_spi
    ].map({y: x for x, y in REGIONS.items()})

    return pd.concat([main, missing_spi]).fillna(0)


def add_id_variables(spi: h5py.File, main: DataFrame):
    spi["person_id"] = main.index
    spi["person_benunit_id"] = main.index
    spi["person_household_id"] = main.index
    spi["benunit_id"] = main.index
    spi["household_id"] = main.index
    spi["role"] = np.array(["adult"] * len(main)).astype("S")
    spi["person_state_role"] = np.array(["citizen"] * len(main)).astype("S")
    spi["state_id"] = np.array([1])
    spi["person_state_id"] = np.array([1] * len(main))


def add_demographics(spi: h5py.File, main: DataFrame):
    LOWER = np.array([0, 16, 25, 35, 45, 55, 65, 75])
    UPPER = np.array([16, 25, 35, 45, 55, 65, 75, 80])
    age_range = main.AGERANGE
    spi["age"] = LOWER[age_range] + np.random.rand(len(main)) * (
        UPPER[age_range] - LOWER[age_range]
    )

    REGIONS = {
        1: "NORTH_EAST",
        2: "NORTH_WEST",
        3: "YORKSHIRE",
        4: "EAST_MIDLANDS",
        5: "WEST_MIDLANDS",
        6: "EAST_OF_ENGLAND",
        7: "LONDON",
        8: "SOUTH_EAST",
        9: "SOUTH_WEST",
        10: "WALES",
        11: "SCOTLAND",
        12: "NORTHERN_IRELAND",
    }

    spi["region"] = np.array(
        [REGIONS.get(x, "UNKNOWN") for x in main.GORCODE]
    ).astype("S")


def add_incomes(spi: h5py.File, main: DataFrame):
    RENAMES = dict(
        pension_income="PENSION",
        self_employment_income="PROFITS",
        property_income="INCPROP",
        savings_interest_income="INCBBS",
        dividend_income="DIVIDENDS",
        blind_persons_allowance="BPADUE",
        married_couples_allowance="MCAS",
        gift_aid="GIFTAID",
        capital_allowances="CAPALL",
        deficiency_relief="DEFICIEN",
        covenanted_payments="COVNTS",
        charitable_investment_gifts="GIFTINV",
        employment_expenses="EPB",
        other_deductions="MOTHDED",
        pension_contributions="PENSRLF",
        person_weight="FACT",
        benunit_weight="FACT",
        household_weight="FACT",
    )
    spi["pays_scottish_income_tax"] = main.SCOT_TXP == 1
    spi["employment_income"] = main[["PAY", "EPB", "TAXTERM"]].sum(axis=1)
    spi["social_security_income"] = main[
        ["SRP", "INCPBEN", "UBISJA", "OSSBEN"]
    ].sum(axis=1)
    spi["miscellaneous_income"] = main[
        ["OTHERINV", "OTHERINC", "MOTHINC"]
    ].sum(axis=1)
    for var, key in RENAMES.items():
        spi[var] = main[key]



from openfisca_uk.data.datasets.frs.enhanced.stages.calibration.losses.loss_category import (
    LossCategory,
)
import tensorflow as tf
from openfisca_uk import Microsimulation
from typing import Iterable, Tuple
from openfisca_uk.parameters import parameters


class BudgetaryImpactByProgramByCountry(LossCategory):
    label = "Country-level aggregates"
    parameter_folder = (
        parameters.calibration.budgetary_impact.by_program_by_country
    )
    weight = 1 / 20

    def get_loss_subcomponents(
        sim: Microsimulation, household_weights: tf.Tensor, year: int
    ) -> Iterable[Tuple]:
        aggregates = BudgetaryImpactByProgramByCountry.parameter_folder
        variables = aggregates.children
        hh_country = sim.calc("country").values
        for variable in variables:
            countries = variables[variable].children
            for country in countries:
                parameter = aggregates.children[variable].children[country]
                in_country = hh_country == country
                values = sim.calc(variable, period=year).values
                entity = sim.simulation.tax_benefit_system.variables[
                    variable
                ].entity.key
                household_totals = sim.map_to(values, entity, "household")
                aggregate = tf.reduce_sum(
                    household_weights * in_country * household_totals
                )
                target = parameter(f"{year}-01-01")
                yield parameter.name + "." + str(year), aggregate, target



from openfisca_tools.data import Dataset, PrivateDataset
from openfisca_uk.data.datasets.frs.enhanced.utils import add_variables
from openfisca_uk.data.storage import OPENFISCA_UK_MICRODATA_FOLDER
from ..calibration import CalibratedFRS
import h5py
from .lcfs_imputation import impute_consumption
from .was_imputation import impute_wealth


class EnhancedFRS(PrivateDataset):
    name = "enhanced_frs"
    label = "Enhanced FRS"
    data_format = Dataset.TIME_PERIOD_ARRAYS
    folder_path = OPENFISCA_UK_MICRODATA_FOLDER
    filename_by_year = {2022: "enhanced_frs_2022.h5"}

    def generate(self, year: int):
        if year not in CalibratedFRS.years:
            ok = input(
                f"Calibrated FRS not found for year {year}. Generate it? (y/n)"
            )
            if ok == "y":
                CalibratedFRS.generate(year)

        # Copy all data from the calibrated FRS

        calibrated_frs = CalibratedFRS.load(year)
        enhanced_frs = h5py.File(self.file(year), mode="w")
        for variable in calibrated_frs.keys():
            for period in calibrated_frs[variable].keys():
                enhanced_frs[f"{variable}/{period}"] = calibrated_frs[
                    variable
                ][period][...]
        calibrated_frs.close()
        enhanced_frs.close()

        pred_consumption = impute_consumption(year, dataset=self)
        add_variables(
            self,
            year,
            {
                f"{field}/{year}": pred_consumption[field]
                for field in pred_consumption.columns
            },
        )

        pred_wealth = impute_wealth(year, dataset=self)
        add_variables(
            self,
            year,
            {
                f"{field}/{year}": pred_wealth[field]
                for field in pred_wealth.columns
            },
        )
        from ..baseline_variables import generate_baseline_variables

        # Import here to avoid circular dependency
        generate_baseline_variables(year)

        from ..remove_zero_weight_households import (
            remove_zero_weight_households,
        )

        remove_zero_weight_households(self, year)


EnhancedFRS = EnhancedFRS()



from typing import Tuple
import pandas as pd
from pathlib import Path
from openfisca_uk.data.datasets.frs.frs import FRS
from openfisca_uk.data.datasets.lcfs import LCFS
from microdf import MicroDataFrame
import synthimpute as si

IMPUTE_VARIABLES = [
    "food_and_non_alcoholic_beverages_consumption",
    "alcohol_and_tobacco_consumption",
    "clothing_and_footwear_consumption",
    "housing_water_and_electricity_consumption",
    "household_furnishings_consumption",
    "health_consumption",
    "transport_consumption",
    "communication_consumption",
    "recreation_consumption",
    "education_consumption",
    "restaurants_and_hotels_consumption",
    "miscellaneous_consumption",
    "petrol_spending",
    "diesel_spending",
]

PREDICTOR_VARIABLES = [
    "is_adult",
    "is_child",
    "region",
    "employment_income",
    "self_employment_income",
    "state_pension",
    "pension_income",
]

REGIONS = {
    1: "NORTH_EAST",
    2: "NORTH_WEST",
    3: "YORKSHIRE",
    4: "EAST_MIDLANDS",
    5: "WEST_MIDLANDS",
    6: "EAST_OF_ENGLAND",
    7: "LONDON",
    8: "SOUTH_EAST",
    9: "SOUTH_WEST",
    10: "WALES",
    11: "SCOTLAND",
    12: "NORTHERN_IRELAND",
}


def impute_consumption(year: int, dataset: type = FRS) -> pd.Series:
    """Impute consumption by fitting a random forest model.

    Args:
        year (int): The year of LCFS to use.
        dataset (type): The dataset to use.

    Returns:
        pd.Series: The imputed consumption categories.
    """

    from openfisca_uk import Microsimulation

    sender = Microsimulation(dataset=LCFS, year=year,).df(
        PREDICTOR_VARIABLES + IMPUTE_VARIABLES,
        map_to="household",
    )

    receiver = Microsimulation(dataset=dataset, year=year,).df(
        PREDICTOR_VARIABLES,
        map_to="household",
    )

    sender.region = sender.region.map(
        {name: float(i) for i, name in REGIONS.items()}
    )
    receiver.region = receiver.region.map(
        {name: float(i) for i, name in REGIONS.items()}
    )

    return si.rf_impute(
        x_train=sender.drop(IMPUTE_VARIABLES, axis=1),
        y_train=sender[IMPUTE_VARIABLES],
        x_new=receiver,
        verbose=True,
    )
